      program p1drcn
c     This is a MAIN PROGRAM for the solution to the Robin boundary
c     value problem in the 1-D case
c
c                  -d(K du/dx)/dx=f,    a < x < b
c
c      -K*(du/dx)(a) + alpa*u(a) = psia, at x=a,
c       K*(du/dx)(b) + alpb*u(b) = psib, at x=b,
c       alpa, alpb > 0.
c
c=================================================================
c    Discretization - SCALAR functions in CELLS ( HC )
c                   - VECTOR functions in NODES ( HN )
c=================================================================
      implicit real*8 (a-h,o-z)
c================================================================
c     eps - constant in the condition of finishing the iteration in 
c     the Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps
c
c     ====================================================================
c     Set of coordinates of grid points
c     call subroutine mkgrid

      call mkgrid


c     Set boundary conditions 
c     call subroutine bcon
      call bcon     

c     Computation of coefficient K in nodes
c     call subroutine coeff
      call coeff

c     Computation of the right-hand side of the differential equation in cells
c     call subroutine right 
      call right


c     computation of coefficients and the right-hand side of the
c     finite-difference equations
      call CFDS

c     Set of constants for the termination of the iteration process
      eps=1.d-12

c     solution for the finite-difference equations by the GAUSS-SIEDEL method
      call siedel 

c     Computing the error
      call error

c--------------------------------------------------------------
c      Test output
c--------------------------------------------------------------
       call output
      end

c**************************************************************

      subroutine CFDS 
c     This is subroutine CFDS for the computation of coefficients
c     and the right-hand side of the finite-difference scheme for 
c     the Robin boundary problem for the Poisson equation
c
c                  -d(K du/dx)/dx=f,    a < x < b
c
c      -K*(du/dx)(a) + alpa*u(a) = psia, at x=a,
c       K*(du/dx)(b) + alpb*u(b) = psib, at x=b,
c       alpa, alpb > 0.
c      for 1-D case
c
c=================================================================
c     Discretization - Scalar functions in Cells ( HC ), 
c                      Vector functions in NODES ( HN )
c=================================================================
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M
c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)
c     ZK - array (1 to M), value of coefficient K in nodes
      common/k/ zk(200)
c     F - array (1 to M-1), values of the right-hand side of the
c                           differential equation in cells
      common/f/ f(200)

c     alpa, alpb - constants in the left-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/alpha/ alpa, alpb

c     psia, psib - constants in the right-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/psi/ psia, psib

c     u -array (0 to M ), solution for the finite-difference scheme.
      common/U/ U(0:200)
c     =================================================================
c     eps - constant in the condition of finishing the iteration in the 
c     Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c

      common/eps/ eps

c     =================================================================
c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 0 to M ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(0:200),c(0:200),e(0:200),phi(0:200)

c     VC, VN - arrays for the volume of cells and nodes
      dimension VC(200), VN(200)

c     Computing VC and VN
      do i=1,M-1
      VC(i) = x(i+1)-x(i)
      end do  

      VN(1)=0.5*(x(2)-x(1))
      do i=2,M-1
      VN(i)=0.5*(x(i+1)-x(i-1))
      end do    
      VN(M)=0.5*(x(M)-x(M-1))
c

c     computing the coefficients of the finite-difference scheme
c     and the right-hand side
c     ========================================================

      c(0) = zk(1)/vn(1) + alpa
      e(0) = -zk(1)/vn(1)
       phi(0) = psia

      do i=1,M-1
      w(i)=-zk(i)/vn(i)
      c(i)=zk(i)/vn(i)+zk(i+1)/vn(i+1)
      e(i)=-zk(i+1)/vn(i+1)
      phi(i)=f(i) * VC(i)
      end do     

      w(M)=-zk(M)/VN(M)
      c(M)=zk(M)/VN(M)+alpb
       phi(M)=psib
       return
       end
c**************************************************************
      
       subroutine siedel
       implicit real*8 (a-h,o-z)
c=============================================================
c     M - number of nodes in the grid
      common/M/ M
c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)
c     ZK - array (1 to M), value of coefficient K in nodes
      common/k/ zk(200)
c     F - array (1 to M-1), values for the right-hand side of the
c                           differential equation in cells
      common/f/ f(200)

c     alpa, alpb - constants in the left-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/alpha/ alpa, alpb

c     psia, psib - constants in the right-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/psi/ psia, psib

c     u -array (0 to M ), solution for the finite-difference scheme.
      common/U/ U(0:200)
c     =================================================================
c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c

      common/eps/ eps

c     =================================================================
c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 0 to M ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(0:200),c(0:200),e(0:200),phi(0:200)

c     =========================================================
c
c     SOLUTION for the system of linear algebraic equations
c     by the GAUSS-SIEDEL iteration method
c
c     SET an INITIAL value for the iteration solution
      do i=0, M
      U(i)=0.             
      end do      

c     beginning of the iteration process

5     continue

      USP1=(phi(0)-e(0)*U(1))/c(0)   
      EPSMAX=dabs(USP1-U(0))
      U(0)=USP1

      do i=1, M-1
      USP1=( phi(i)-w(i)*U(i-1)-e(i)*U(i+1))/c(i)
      EPSMAX=amax1(EPSMAX,dabs(USP1 - U(i)))
      U(i)=USP1
      end do   

      USP1=(phi(M)-w(M)*U(M-1))/c(M)
      EPSMAX=amax1(EPSMAX,dabs(USP1-U(M)))
      U(M)=USP1

c     checking the termination condition
      if(EPSMAX.GT.EPS) goto 5
      return
      end
c*****************************************************

      subroutine mkgrid

      implicit real*8 (a-h,o-z)
c     This subroutine generates a grid in 1-D
c     That is, it generates array X (1 to M) of
c     coordinates of the grid nodes

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     Set number of nodes
      M=17      

c     Set end points of the computational domain
      a=0.d0
      b=1.d0

c     Simple smooth grid as in the examples of the book

      do i=1,m
      xi=a+(b-a)*(i-1.d0)/(M-1)
      x(i)=xi**2
      end do
      return
      end
c************************************************************
      subroutine exact
      implicit real*8 (a-h,o-z)
c     Compute the exact soulution in the nodes
c     Results - array uex (1 to M)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     UEX - array (0 to M) exact solution in the nodes
      common/uex/ uex(0:200)
      
       pi=acos(-1.)
      

       uex(0)=sin(2*pi*x(1))
      do i=1,M-1
       uex(i)=sin(2*pi*0.5*(x(i)+x(i+1)))
      end do  
       uex(M)=sin(2*pi*x(M))
      return 
      end
c************************************************************
       subroutine output
       implicit real*8 (a-h,o-z)
c      Output subroutine 

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     u -array (0 to M ), solution for the finite-difference scheme.
      common/U/ U(0:200)

c     UEX - array (0 to M) exact solution in the nodes
      common/uex/ uex(0:200)

c     ERR - array (0 to M), error between the exact
c     and approximate solutions
      common/err/ERR(0:200)

c      Max-norm and L2-norm of the error between
c      the exact and approximate solutions 
       common/errors/ supn, zL2N

c       Open the file p1drnc.output for
c       test output of program P1DDNC
       open(unit=1,file='p1drcn.output')

       write(1,33)
33     format(15X, 'TEST OUTPUT FOR PROGRAM p1drcn')
c     =========================================================
      write(1,*)
      write(1,8) M, supn, zL2N
8     format(2X,'Number of Nodes=',i5, 2X,
     1 'MAX-NORM = ', E12.5, 2X,'  L2-NORM = ', E12.5)
       write(1,*)

c     Print of the approximate and exact solutions of the test problem
       write(1,9) 
9      format(5X,'i',7X,' X',10X,
     1 'Appr.Sol.',4X,'Exact.Sol.',5X,'Error')
       write(1,7) 0, x(1),u(0), uex(0), err(0)
       do i=1,M-1
       xci=0.5*(x(i)+x(i+1))
       write(1,7) i, xci,u(i), uex(i), err(i)
       end do  
       write(1,7) M, x(M),u(M), uex(M), err(M)
 7     format(1X, i5,2X, E12.5,2X, E12.5,2X, E12.5,2X, E12.5)
       return
       end
c**************************************************************
      subroutine error
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     VC - array ( 1 to M-1 ) volume of cells
c     COMMENTS - this volume is needed only for computing
c     the L2 Norm of error
      dimension VC(200)

c     u -array (0 to M ), solution for the finite-difference scheme.
      common/U/ U(0:200)

c     UEX - array (0 to M) exact solution in the nodes
      common/uex/ uex(0:200)

c      Sup-norm and L2-norm of the error between
c      the exact and approximate solutions
       common/errors/ supn, zL2N

c     ERR - array (0 to M), error between the exact
c     and approximate solutions
      common/err/ERR(0:200)

c      Computing the exact solution
       call exact

c     Computation of the volume of cells
      do i=1,m-1
      VC(i) = x(i+1) - x(i)
      end do     

c     Error
c     -----
      do i=0,M
      err(i)= uex(i)-u(i)
      end do     

c     sup - norm
c     -----------
      supn = 0.d0
      do i=0,M
      supn=amax1(supn,abs(err(i)))
      end do      

c     L  - norm
c      2
c     ----------

      zL2N = 0.d0
      do i=1, M-1
      zL2N = zL2N + (err(i))**2*VC(i)
      end do  
      zL2N=zL2N+err(0)**2+err(M)**2   
      zL2N = dsqrt(zL2N)
      return
      end
c*********************************************************************
      subroutine bcon
      implicit real*8 (a-h,o-z)

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     psia, psib - constants in the Dirichlet boundary conditions in 
c     points a and b respectively.
      common/psi/ psia, psib

c     alpa, alpb - coefficients near u(a) and u(b)
c     in Robin boundary conditions
      common/alpha/ alpa, alpb

      pi=acos(-1.)

      psia=-2.*pi
      psib=4.*pi

      alpa=1.
      alpb=1.

      return
      end
c*********************************************************************

      subroutine coeff
c     In this subroutine, the coefficients of the finite-difference
c     scheme in nodes must be computed
      implicit real*8 (a-h,o-z)


c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     ZK - array (1 to M-1), value of coefficient K in nodes
      common/k/ zk(200)


      do i=1,m
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
      zk(i)=fk(x(i))
      end do 
      return
      end
c****************************************************************

      real function fk*8(x)
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
     
      implicit real*8 (a-h,o-z)
      fk=x**4 + 1.
      return
      end
c****************************************************************

      subroutine right
c     This subroutine computes the values of the right-hand side
c     of the original differential equation in cells

      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)


c     F - array (1 to M-1), values for the right-hand side of the
c                           differential equation in cells
      common/f/ f(200)



      do i=1,m-1  
       f(i)=fr(0.5*(x(i)+x(i+1)))
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x
      end do
      return 
      end
c*********************************************************************
      
      real function fr*8 (x)
      implicit real*8 (a-h,o-z)
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x

c     Number pi=3.14...
      pi=acos(-1.)
       fr=-2*pi*4.d0*x**3*cos(2*pi*x)
     1      +4*pi**2*(x**4+1)*sin(2*pi*x)
       return
       end
c**********************************************************************


