c     program p1drnc
c     This is a MAIN PROGRAM for the solution to the Robin boundary
c     value problem in the 1-D case
c
c                  -d(K du/dx)/dx=f,    a < x < b
c
c      -K*(du/dx)(a) + alpa*u(a) = psia, at x=a,
c       K*(du/dx)(b) + alpb*u(b) = psib, at x=b,
c       alpa, alpb > 0.
c
c=================================================================
c    Discretization - SCALAR functions in NODES ( HN )
c                   - VECTOR functions in CELLS ( HC )
c=================================================================
      implicit real*8 (a-h,o-z)
c================================================================
c     eps - constant in the condition of finishing the iteration in 
c     the Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps
c
c     IAPP - order of approximation for the Robin boundary
c            conditions
      common/iapp/ IAPP
c     ====================================================================
c     Set of coordinates of grid points
c     call subroutine mkgrid

      call mkgrid


c     Set boundary conditions 
c     call subroutine bcon
      call bcon     

c     Computation of coefficient K in cells
c     call subroutine coeff
      call coeff

c     Computation of the right-hand side of the differential equation in nodes
c     call subroutine right 
      call right


c     computation of coefficients and the right-hand side of the
c     finite-difference equations
c     IAPP - order of approximation of Robin boundary conditions
c     IAPP=1 - first order
c     IAPP=2 - second order
      IAPP=1
      call CFDS(IAPP)

c     Set of constants for the termination of the iteration process
      eps=1.d-12

c     solution for finite-difference equations by the GAUSS-SIEDEL method
      call siedel 

c     Computing the error
      call error

c--------------------------------------------------------------
c      Test output
c--------------------------------------------------------------
       call output
      end

c**************************************************************

      subroutine CFDS(IAPP) 
c     This is subroutine CFDS for the computation of coefficients
c     and the right-hand side of the finite-difference scheme for the
c     Robin boundary problem for the Poisson equation
c
c                  -d(K du/dx)/dx=f,    a < x < b
c
c      -K*(du/dx)(a) + alpa*u(a) = psia, at x=a,
c       K*(du/dx)(b) + alpb*u(b) = psib, at x=b,
c       alpa, alpb > 0.
c      for 1-D case
c
c=================================================================
c     Discretization - Scalar functions in Nodes ( HN ), 
c                      Vector functions in Cells ( HC )
c=================================================================
c     Parameter IAPP determines the order of approximation for the
c     Robin boundary conditions
c     IAPP=1 - first order,
c     IAPP=2 -second order.
c=================================================================
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M
c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)
c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)
c     F - array (2 to M-1), values for the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)

c     alpa, alpb - constants in the left-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/alpha/ alpa, alpb

c     psia, psib - constants in the right-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/psi/ psia, psib

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)
c     =================================================================
c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration procedure. The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c

      common/eps/ eps

c     =================================================================
c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 1 to M ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)

c     VC, VN - arrays for the volume of cells and nodes
      dimension VC(200), VN(200)

c     Computing VC and VN
      do i=1,M-1
      VC(i) = x(i+1)-x(i)
      end do  

      VN(1)=0.5*(x(2)-x(1))
      do i=2,M-1
      VN(i)=0.5*(x(i+1)-x(i-1))
      end do    
      VN(M)=0.5*(x(M)-x(M-1))
c

c     computing the coefficients of the finite-difference scheme
c     and the right-hand side
c     ========================================================

      c(1) = zk(1)/vc(1) + alpa
      e(1) = -zk(1)/vc(1)
      if(IAPP.eq.1) then
c     first order approximation for the Robin boundary condition
       phi(1) = psia
       else
c     second order approximation for the Robin boundary condition
       phi(1) = psia + VN(1)*f(1)
       end if

      do i=2,M-1
      w(i)=-zk(i-1)/vc(i-1)
      c(i)=zk(i-1)/vc(i-1)+zk(i)/vc(i)
      e(i)=-zk(i)/vc(i)
      phi(i)=f(i) * VN(i)
      end do     

      w(M)=-zk(M-1)/VC(M-1)
      c(M)=zk(M-1)/VC(M-1)+alpb
      if(IAPP.eq.1) then
c     first order approximation for the Robin boundary condition
       phi(M)=psib
       else
c     second order approximation for the Robin boundary condition
       phi(M)=psib + VN(M) * f(M)
       end if

       return
       end
c**************************************************************
      
       subroutine siedel
       implicit real*8 (a-h,o-z)
c=============================================================
c     M - number of nodes in the grid
      common/M/ M
c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)
c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)
c     F - array (2 to M-1), values for the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)

c     alpa, alpb - constants in the left-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/alpha/ alpa, alpb

c     psia, psib - constants in the right-hand side of the Robin 
c     boundary conditions in points a and b respectively.
      common/psi/ psia, psib

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)
c     =================================================================
c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration prosedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c

      common/eps/ eps

c     =================================================================
c     W, C, E, PHI - arrays for the coefficients and the right-hand side
c     ( 1 to M ) in the finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)

c     =========================================================
c
c     SOLUTION for the system of linear algebraic equations
c     by the GAUSS-SIEDEL iteration method
c
c     SET an INITIAL value for the iteration solution
      pi=acos(-1.)
      do i=1, M
      U(i)=sin(2*pi*x(i))
      end do      

c     beginning of the iteration process

5     continue

      USP1=(phi(1)-e(1)*U(2))/c(1)   
      EPSMAX=dabs(USP1-U(1))
      U(1)=USP1

      do i=2, M-1
      USP1=( phi(i)-w(i)*U(i-1)-e(i)*U(i+1))/c(i)
      EPSMAX=amax1(EPSMAX,dabs(USP1 - U(i)))
      U(i)=USP1
      end do   

      USP1=(phi(M)-w(M)*U(M-1))/c(M)
      EPSMAX=amax1(EPSMAX,dabs(USP1-U(M)))
      U(M)=USP1

c     checking the termination condition
      if(EPSMAX.GT.EPS) goto 5
      return
      end
c*****************************************************

      subroutine mkgrid

      implicit real*8 (a-h,o-z)
c     This subroutine generates a grid in 1-D
c     That is, it generates array X (1 to M) of
c     coordinates of the grid nodes

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     Set number of nodes
      M=17      

c     Set end points of the computational domain
      a=0.d0
      b=1.d0

c     Simple smooth grid as in the examples of the book

      do i=1,m
      xi=a+(b-a)*(i-1.d0)/(M-1)
      x(i)=xi**2
      end do
      return
      end
c************************************************************
      subroutine exact
      implicit real*8 (a-h,o-z)
c     Compute the exact soulution in the nodes
c     Results - array uex (1 to M)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     UEX - array (1 to M) exact solution in the nodes
      common/uex/ uex(200)
      
       pi=acos(-1.)

      do i=1,M
       uex(i)=sin(2*pi*x(i))
      end do  
      return 
      end
c************************************************************
       subroutine output
       implicit real*8 (a-h,o-z)
c      Output subroutine 

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M) exact solution in the nodes
      common/uex/ uex(200)

c     IAPP - order of approximation for the Robin boundary
c            conditions
      common/iapp/ IAPP

c     ERR - array (1 to M), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c      Max-norm and L2-norm of the error between
c      the exact and approximate solutions 
       common/errors/ supn, zL2N

c       Open the file p1drnc.output for
c       test output of program P1DDNC
       open(unit=1,file='p1drnc.output')

       write(1,33)
33     format(15X, 'TEST OUTPUT FOR PROGRAM p1drnc')
c     =========================================================
      write(1,*)
      if(IAPP.eq.1) then
      write(1,44)
      else
      write(1,55)
      end if
44    format(7X, 'FIRST ORDER APPROXIMATION OF BOUNDARY CONDITIONS')
55    format(7X, 'SECOND ORDER APPROXIMATION OF BOUNDARY CONDITIONS')
      write(1,*)

      write(1,8) M, supn, zL2N
8     format(2X,'Number of Nodes=',i5, 2X,
     1 'MAX-NORM = ', E12.5, 2X,'  L2-NORM = ', E12.5)
       write(1,*)

c     Print of the approximate and exact solutions of the test problem
       write(1,9) 
9      format(5X,'i',7X,' X',10X,
     1 'Appr.Sol.',4X,'Exact.Sol.',5X,'Error')
       do i=1,M
       write(1,7) i, x(i), u(i), uex(i), err(i)
       end do  
 7     format(1X, i5,2X, E12.5,2X, E12.5,2X, E12.5,2X, E12.5)
       return
       end
c**************************************************************
      subroutine error
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     VN - array ( 1 to M ) volume of nodes
c     COMMENTS - this volume is needed only for computing
c     the L2 Norm of error
      dimension VN(200)

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M) exact solution in the nodes
      common/uex/ uex(200)

c      Sup-norm and L2-norm of the error between
c      the exact and approximate solutions
       common/errors/ supn, zL2N

c     ERR - array (1 to M), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c      Computing the exact solution
       call exact

c     Computation of the volumes of nodes
      VN(1) = 0.5 * (x(2) - x(1))
      do i=2,m-1
      VN(i) = 0.5 * (x(i+1) - x(i-1))
      end do     
      VN(M) = 0.5 * (x(M) - x(M-1))

c     Error
c     -----
      do i=1,M
      err(i)= uex(i)-u(i)
      end do     

c     sup - norm
c     -----------
      supn = 0.d0
      do i=1,M
      supn=amax1(supn,abs(err(i)))
      end do      

c     L  - norm
c      2
c     ----------

      zL2N = 0.d0
      do i=2, M-1
      zL2N = zL2N + (err(i))**2*VN(i)
      end do  
      zL2N=zL2N+err(1)**2+err(M)**2   
      zL2N = dsqrt(zL2N)
      return
      end
c*********************************************************************
      subroutine bcon
      implicit real*8 (a-h,o-z)

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     psia, psib - constants in the Dirichlet boundary conditions in 
c     points a and b respectively.
      common/psi/ psia, psib

c     alpa, alpb - coefficients near u(a) and u(b)
c     in the Robin boundary conditions
      common/alpha/ alpa, alpb

      pi=acos(-1.)

      psia=-2.*pi
      psib=4.*pi

      alpa=1.
      alpb=1.

      return
      end
c*********************************************************************

      subroutine coeff
c     In this subroutine, the coefficients of the finite-difference
c     scheme in cells must be computed
      implicit real*8 (a-h,o-z)


c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)


      do i=1,m-1
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
      zk(i)=fk(0.5*(x(i+1)+x(i)))
      end do 
      return
      end
c****************************************************************

      real function fk*8(x)
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
     
      implicit real*8 (a-h,o-z)
      fk=x**4 + 1.
      return
      end
c****************************************************************

      subroutine right
c     This subroutine computes the values of the right-hand side
c     of the original differential equation in nodes

      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)


c     F - array (2 to M-1), values for the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)



      do i=1,m  
       f(i)=fr(x(i))
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x
      end do
      return 
      end
c*********************************************************************
      
      real function fr*8 (x)
      implicit real*8 (a-h,o-z)
c      fr(x) - function which returns the values of the right-hand side
c              of the original differential equation in point x

c     Number pi=3.14...
      pi=acos(-1.)
       fr=-2*pi*4.d0*x**3*cos(2*pi*x)
     1      +4*pi**2*(x**4+1)*sin(2*pi*x)
       return
       end
c**********************************************************************


