        program h1ddnc
c     This is a MAIN PROGRAM for the solution to the Dirichlet boundary
c     value problem for Heat equations
c            du/dt = d(K du/dx)/dx+f(x,t)
c             u(a,t)=psia(t), u(b,t)=psib(t)
c      for 1-D case.
c      K=K(x,t), does not depend on time.
c=================================================================
c    Discretization - SCALAR functions in NODES ( HN )
c                   - VECTOR functions in CELLS ( HC ) 
c=================================================================
      implicit real*8 (a-h,o-z)
c============================================================================
c
c     t - current time  and time step 
      common/t/ t, dt

c     eps - constant in the condition of finishing the iteration in 
c     the Gauss-Siedel iteration procedure. The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps

c    sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
c======================================================================
c
c     Set maximal time
      timemax = 0.1
 
c     Set interval of time for output
      dtout=0.045 
c     Set first time moment for output
      tout=dtout
     
c     Set up time step and initial time
       t=0.
       dt = 0.00004      

c     Set type of finite-difference scheme
c     explicit  sigma=0. or implicit  sigma is not equal to zero

c     Explicit scheme
c     sigma=0.

c     Implicit scheme
      sigma=0.7 

c     Set of coordinates for grid points
c     call subroutine mkgrid
c     (grid does not depend on time)

      call mkgrid
 
c     Computation of geometrical quantities,
c     which do not depend on time
 
      call geometry

c     Set initial conditions
      call initial        

c     Beginning of time integration

1111   continue
        t = t + dt
c     Set boundary conditions 
c     call subroutine bcon
      call bcon     

c     Computation of coefficient K of the original equation in cells
      call coeff

c     Computation of the right-hand side of the differential equation in nodes
      call right

c     Computation of approximation for the right-hand side of the original
c     differential equation
      call sphi

c      computation of coefficient for the elliptic operator
c      which is needed for explicit and implicit schemes.
      call ellcoef

c     Check Explicit or Implicit Scheme
      if(sigma.ne.0) then

c     Implicit scheme
c     computation of coefficients and the right-hand side of the
c     finite-difference equations for the case of implicit schemes
      call hcoeff
 
c     Set of constants for the termination of the iteration process
      eps=1.d-8  

c     solution for finite-difference equations by the GAUSS-SIEDEL method
      call siedel 

      
      else

c     Explicit scheme
      call step
     
      end if

c--------------------------------------------------------------
c      Test output
c--------------------------------------------------------------
      if(t.ge.tout) then
      write(*,*) t,tout
c     Set next output time
       tout=tout+dtout

c     Computing the error
      call error
       call output
      end if
        if(t.gt.timemax) stop 
c      Go to computation of the new time level

       goto 1111
      end

c**************************************************************

      subroutine ellcoef
      implicit real*8 (a-h,o-z)
c     This is a subroutine ellcoeff for the computation of coefficients
c     discrete analog of the general elliptic operator in 1D      
c                    -d(K du/dx)/dx
c=================================================================
c     Discretization - Scalar functions in Nodes ( HN ),
c                      Vector functions in Cells  ( HC )
c=================================================================
      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M


c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)

c     =================================================================
c     W, C, E, PHI - arrays for coefficients and the right-hand side
c     ( 2 to M-1 ) in finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)
c     =================================================================

c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200), VN(200)
c

c     computing the coefficients of the finite-difference scheme
c     and the right-hand side
c     ========================================================


      do 3 i=2,M-1
      w(i)=(-zk(i-1)/vc(i-1))/VN(i)
      c(i)=(zk(i-1)/vc(i-1)+zk(i)/vc(i))/VN(i)
      e(i)=(-zk(i)/vc(i))/VN(i)
3     continue

      return
      end
c*************************************************************

      subroutine siedel
      implicit real*8 (a-h,o-z)
c
c     SOLUTION for the system of linear algebraic equations
c     by the GAUSS-SIEDEL iteration method
c
c     M - number of nodes in the grid
      common/M/ M

c     U -array (1 to M ), solution for the finite-difference scheme.
      common/U/ u(200)

c     W, C, E, PHI - arrays for coefficients and the right-hand side
c     ( 2 to M-1 ) in finite-difference equations
c
c                 W(i)*U(i-1) + C(i)*U(i) + E(i)*U(i+1) = PFI(i)
c
      common/coefds/ w(200),c(200),e(200),phi(200)

c     psia, psib - constants in the Dirichlet boundary conditions in
c     points a and b respectively.
      common/psi/ psia, psib

c     eps - constant in the condition of finishing the iteration in 
c     the Gauss-Siedel iteration procedure.  The termination condition is
c
c              (S+1)  (S)
c     max( abs( U(i) - U(i))) < eps
c
c     where S is the iteration number.
c
      common/eps/ eps
c============================================================================
      U(1)=psia
      U(M)=psib

c     Initial guess is the values from the previous time step
c     which is in array U
   
c     beginning of the iteration process
5     continue

      EPSMAX=0.

      do 6 i=2, M-1
      USP1=( phi(i)-w(i)*U(i-1)-e(i)*U(i+1))/c(i)
      EPSMAX=amax1(EPSMAX,dabs(USP1 - U(i)))
      U(i)=USP1
6     continue

c     checking the termination condition
      if(EPSMAX.GT.EPS) goto 5
      return
      end
c************************************************************

      subroutine mkgrid
      implicit real*8 (a-h,o-z)
c     This subroutine generates a grid in 1-D
c     That is, it generates array X (1 to M) of
c     the coordinates of grid nodes

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     Set number of nodes
      M=17      

c     Set end points of the computational domain
      a=0.d0
      b=1.d0

c     Simple smooth grid as in the examples of the book

      do i=1,m
      xi=a+(b-a)*(i-1.d0)/(M-1)
      x(i)=xi**2
      end do

      return
      end
c************************************************************

      subroutine exact
      implicit real*8 (a-h,o-z)
c     Compute the exact solution in the nodes
c     Results - array uex (1 to M)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     UEX - array (1 to M) exact solution in the nodes
c           for current time t
      common/uex/ uex(200)
      
c     t - current time  
      common/t/ t, dt

       pi=acos(-1.)

      do i=1,M
       uex(i)=exp(-4.*pi**2*t)*sin(2*pi*x(i))
      end do  
      return 
      end
c************************************************************

       subroutine output
       implicit real*8 (a-h,o-z)
c      Output subroutine 

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M) exact solution in the nodes
      common/uex/ uex(200)

c     ERR - array (1 to M), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c     t - current time  
      common/t/ t, dt

c      Max-norm and L2-norm of the error between
c      the exact and approximate solutions 
       common/errors/ supn, zL2N

      common/expimp/sigma

c       Open the file h1ddnc.output for
c       test output of program H1DDNC
       open(unit=1,file='h1ddnc.output')

       write(1,33)
33     format(15X, 'TEST OUTPUT FOR PROGRAM h1ddnc')
c     =========================================================
      write(1,*)
      if(sigma.eq.0) then
      write(1,*)'               Explicit Scheme'
      else
      write(1,*)'               Implicit Scheme',' Sigma= ', sigma
      end if
      write(1,*)


      write(1,8) t,dt, M
      write(1,*)
      write(1,88)  supn, zL2N
8     format(2X,'Time= ',E12.5,2X,'dt = ',E12.5,2X,
     *'Number of Nodes=',i5)
88     format(2X,
     1 'MAX-NORM = ', E12.5, 2X,'  L2-NORM = ', E12.5)
       write(1,*)

c     Print of the approximate and exact solutions of the test problem
       write(1,9) 
9      format(5X,'i',7X,' X',10X,
     1 'Appr.Sol.',4X,'Exact.Sol.',5X,'Error')
       do i=1,M
       write(1,7) i, x(i), u(i), uex(i), err(i)
       end do  
      write(1,*)
 7     format(1X, i5,2X, E12.5,2X, E12.5,2X, E12.5,2X, E12.5)
77     format(1X,  E12.5,2X, E12.5)
       return
       end
c**************************************************************

      subroutine error
      implicit real*8 (a-h,o-z)
c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     VN - array ( 1 to M ) volume of nodes
c     COMMENTS - this volume is needed only for computing
c     the L2 Norm of error
c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200), VN(200)

c     u -array (1 to M ), solution for the finite-difference scheme.
      common/U/ U(200)

c     UEX - array (1 to M) exact solution in the nodes
      common/uex/ uex(200)

c      Sup-norm and L2-norm of the error between
c      the exact and approximate solutions 
       common/errors/ supn, zL2N

c     ERR - array (1 to M), error between the exact
c     and approximate solutions
      common/err/ERR(200)

c      Computing the exact solution
       call exact

c     Error
c     -----
      do i=1,M
      err(i)= uex(i)-u(i)
      end do     

c     sup - norm
c     -----------
      supn = 0.d0
      do i=2,M-1
      supn=amax1(supn,abs(err(i)))
      end do      

c     L  - norm
c      2
c     ----------

      zL2N = 0.d0
      do i=2, M-1
      zL2N = zL2N + (err(i))**2*VN(i)
      end do     
      zL2N = dsqrt(zL2N)
      return
      end
c****************************************************************
      
      subroutine bcon
      implicit real*8 (a-h,o-z)

c     t - current time  
      common/t/ t, dt

c     a and b coordinates of the ends of the computational domain [a,b]
      common/ab/ a,b

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

      common/psi/ psia, psib
c     psia, psib - determines the Dirichlet boundary conditions in 
c     points a and b respectively to time = t.
      psia=fmu0(t)
      psib=fmu1(t)
      return
      end
c*********************************************************************

      subroutine coeff
c     In this subroutine, the coefficients of the finite-difference
c     scheme in cells must be computed
      implicit real*8 (a-h,o-z)
 
      common/t/ t,dt

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)

 
      tt=t-0.5*dt


      do i=1,m-1
c     fk(x) - is a function which returns the values of coefficient
c     K(x) of the original differential equation
      zk(i)=fk(0.5*(x(i+1)+x(i)),tt)
      end do 
      return
      end
c****************************************************************

      real function fk*8(x,tt)
c     fk(x,tt) - is a function which returns the values of coefficient
c     K(x,tt) of the original differential equation
     
      implicit real*8 (a-h,o-z)
      fk=x**4 + 1.
      return
      end
c****************************************************************

      subroutine right
c     This subroutine computes the values of the right-hand side
c     of the original differential equation in nodes

      implicit real*8 (a-h,o-z)

      common/t/ t,dt

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)


c     F - array (2 to M-1), values of the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)


      tt=t-0.5*dt

      do i=2,m-1
       f(i)=fr(x(i),tt)
c      fr(x,tt) - function which returns the values of the right-hand side
c              of the original differential equation in point x
      end do
      return 
      end
c*********************************************************************
      
      real function fr*8 (x,tt)
      implicit real*8 (a-h,o-z)
c      fr(x,tt) - function which returns the values of the right-hand side
c              of the original differential equation in point x
c              at time moment tt
c     Number pi=3.14...
      pi=acos(-1.)
      fr=exp(-4.*pi**2*tt)*
     * (-4.*pi**2*sin(2.*pi*x)*(1.-
     * (x**4+1.))-4.*x**3*2.*pi*cos(2.*pi*x))
c      fr =0.
       return
       end
c**********************************************************************

      subroutine initial
      implicit real*8 (a-h,o-z)
c     Compute the initial values in the nodes
c     Results - array u (1 to M)

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     U -array (1 to M ), solution for the finite-difference scheme.
      common/U/ u(200)

c     t - current time  
      common/t/ t, dt
       pi=acos(-1.)

c     Set up initial conditions  
      t=0.

      do i=1,M
       u(i)=sin(2*pi*x(i))
      end do  

      return 
      end

************************************************************

      subroutine step
      implicit real*8 (a-h,o-z)
c     This is a subroutine step for the computation of solution for the
c     explicit finite-difference scheme for the Dirichlet
c     boundary problem for Heat equations 
c             du/dt = d(K du/dx)/dx+f(x,t)
c              u(a,t)=zmu0(t), u(b,t)=zmu1(t)
c      for the 1-D case.
c=================================================================
c     Discretization - Scalar functions in Nodes ( HN ),
c                      Vector functions in Cells  ( HC )
c=================================================================
      implicit real*8 (a-h,o-z)

c     t,dt- current time, and time step  
      common/t/ t,dt

c     M - number of nodes in the grid
      common/M/ M

c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     unp1 - array (1 to M), solution on the new time level
c     U    - array (1 to M), solution on the previous time level
      common/U/  U(200)
      dimension unp1(200)

c     ZK - array (1 to M-1), value of coefficient K in cells
      common/k/ zk(200)

c     F - array (2 to M-1), values of the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)

c     psia, psib - value of functions zmu0, zmu1 at moment t_n 
c     in Dirichlet boundary conditions in
c     points a and b respectively.
      common/psi/ psia, psib

c     =================================================================
c     W, C, E, PHI -  coefficients and the right-hand side
c     in the finite-difference equations in node i
c
c                 W*U(i-1) + C*U(i) + E*U(i+1) = PFI
      common/coefds/ w(200),c(200),e(200),phi(200)
c
c     =================================================================

       U(1)=psia
       U(M)=psib
       do i=2,M-1
c      Solution on new time level
       unp1(i) = U(i) 
     * + dt*(-(w(i)*U(i-1)+c(i)*U(i)+e(i)*U(i+1))
     * + phi(i))
        end do
        
c      boundary conditions
       unp1(1)=psia
       unp1(M)=psib

c      Overdetermination of array U,
c      because we use only two arrays.
c      If you need to use both arrays, do it before this 
c      overdetermination and put this loop in another subroutine
       do i=1,M
       U(i) = unp1(i)
       end do

      return
      end
c*************************************************************

      real function fmu0*8 (t)
      implicit real*8 (a-h,o-z)
c      fmu0(t) - function which returns the values of the right-hand side
c      in Dirichlet boundary conditions of the original differential
c      problem, at the end point x=a.  

c     Number pi=3.14...
      pi=acos(-1.)
       fmu0 = 0.
       return
       end
c*************************************************************
      real function fmu1*8 (t)
      implicit real*8 (a-h,o-z)
c      fmu1(t) - function which returns the values of the right-hand side
c      in Dirichlet boundary conditions of the original differential
c      problem, at the end point x=b.

c     Number pi=3.14...
      pi=acos(-1.)
       fmu1 = 0.
       return
       end
c*************************************************************
        subroutine geometry
c       This subroutine computes the volumes of the cells and nodes
c       which are not dependent on time
 
      implicit real*8 (a-h,o-z)
 
c     M - number of nodes in the grid
      common/M/ M
 
c     X - array (1 to M ), coordinates of grid nodes
      common/x/ x(200)

c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200), VN(200)
c     =================================================================

c     Computing VC and VN

      do 1 i=1,M-1
      VC(i) = x(i+1)-x(i)
1     continue

      VN(1)=0.5*(x(2)-x(1))
      do 2 i=2,M-1
      VN(i)=0.5*(x(i+1)-x(i-1))
2     continue
      VN(M)=0.5*(x(M)-x(M-1))

      return 
      end
c******************************************************************
      subroutine hcoeff
      implicit real*8 (a-h,o-z)

c     t,dt- current time, and time step  
      common/t/ t,dt

c    sigma - parameter in the finite-difference scheme
      common/expimp/ sigma

c     M - number of nodes in the grid
      common/M/ M


c     U    - array (1 to M), solution on the previous time level
      common/U/  U(200)


c     F - array (2 to M-1), values of the right-hand side of the
c                           differential equation in nodes
      common/f/ f(200)

c     =================================================================
c     W, C, E, PHI -  coefficients and the right-hand side
c     in the finite-difference equations in node i
c
c                 W*U(i-1) + C*U(i) + E*U(i+1) = PFI
c
      Common/coefds/ w(200),c(200),e(200),phi(200)
c     =================================================================
c     Coefficients and the right-hand side of 
c     of the implicit finite-difference scheme
c     for heat equations

      do i=2,M-1
      phi(i)=phi(i) + U(i)/dt 
     * - (1.-sigma)*(w(i)*U(i-1)+C(i)*U(i)+e(i)*U(i+1))
      w(i)=sigma*w(i)
      c(i)=sigma*c(i) + 1./dt
      e(i)=sigma*e(i)
      end do

      return
      end
c **********************************************************

      subroutine sphi
      implicit real*8 (a-h,o-z)

c     M - number of nodes in the grid
      common/M/ M

      common/f/ f(200)

c     =================================================================
c     W, C, E, PHI -  coefficients and the right-hand side
c     in the finite-difference equations in node i
c
c                 W*U(i-1) + C*U(i) + E*U(i+1) = PFI
c
c     =================================================================

      Common/coefds/ w(200),c(200),e(200),phi(200)

c     =================================================================
c     Coefficients and the right-hand side of 
c     of the implicit finite-difference scheme
c     for heat equations

      do i=2,M-1
      phi(i)=f(i)
      end do

      return
      end
c ********************************************************************


