      program h2ddcn
c     This is a main program for the solution to the Dirichlet boundary
c     value problem for the Heat equation
c    
c         du/dt=[d(KXX du/dx)/dx+d(KXY du/dy)/dx
c               +d(KXY du/dx)/dy+d(KYY du/dy)/dy+f
c                u(x,y,t)=psi(x,y,t) for (x,y) on the boundary,
c             and intial  conditions
c             u(x,y,0)=u_0(x,y) at t=0.

c      for 2-D case.
c=================================================================
c     Discretization - SCALAR functions in CELLS (HC),
c                    - VECTOR functions in NODES (HN)
c=================================================================
      implicit real*8 (a-h,o-z)
c======================================================================
c     eps - constant in the condition of finishing the iteration in the
c     Gauss-Siedel iteration procedure. The termination condition is
c
c             (S+1)     (S)
c     max(abs( U(i,j) - U(i,j))) < eps
c     i,j
c 
c     where S is the iteration number.
c

      common/eps/ eps
c=====================================================================
c     t - current time
      common/t/ t,dt
c=====================================================================
c    sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
c======================================================================
c
c     Set interval of output and maxtime
      dtout=0.01
      timemax = 0.01
 
c     Set first time moment for output
      tout=dtout
 
c     Set initial time moment t=0, and dt.
      t=0.
 
      dt=0.00001
c     Set type of finite-difference scheme:
c     Explicit sigma=0.
c     Implicit sigma <> 0.
 
c     Explicit scheme
c     sigma=0.
 
c     Implicit scheme
      sigma=0.7
 
c     Set of coordinates of grid points
c     (grid does not depend on time)
 
      call mkgrid

c     Computation of geometrical quantities,
c     which do not depend on time
 
      call geometry
 
c     Set initial conditions
 
      call initial
 
c     Beginning of time loop
 
1111  continue
 
c     Set up current time
      t=t+dt

c     Set a boundary condition
c     call subroutine bcon

      call bcon


c     Computation of coefficient K in cells
c     call subroutine coeff

      call coeff

c     Computation of the right-hand side of the differential equation in nodes
c     call subroutine right

      call right

c      computation of the coefficient of the elliptic operator
c      which is needed for explicit and implicit schemes.
 
      call ellcoef

 
      call sphi

c     Check Explicit or Implicit Scheme
      if(sigma.ne.0) then
c    Implicit scheme
c     Computation of coefficients and the right-hand side of the 
c     finite-difference equations
 
      call hcoeff

c     Set of constant for the termination of the iteration process
      eps=1.d-12

c     Solution for the finite-difference equations by the GAUSS-SIEDEL method
c     call subroutine siedel
      call siedel
 
       else
 
c      Explicit scheme
c      subroutine step
       call step
 
       end if
 
c     Test output
c-------------------------------------------------------------------------
      if(t.ge.tout) then
c     Set next output time
 
      tout=tout+dtout
      write(*,*) t,timemax
c     Computing the error
c     subroutine error
 
      call error
 
c     Test output
 
      call output
 
      end if
 
      if(t.gt.timemax) stop
 
c      Go to computation of the new time level
       goto 1111
      end
c*************************************************************************
      subroutine ellcoef
c     This is subroutine ellcoef for the computation of coefficents
c     of the discrete analog of the general elliptic operator      
c
c     -[d(KXX du/dx)/dx+d(KXY du/dy)/dx+d(KXY du/dx)/dy+d(KYY du/dy)/dy
c=================================================================
c     Discretization - SCALAR functions in CELLS (HC),
c                    - VECTOR functions in NODES (HN)
c=================================================================
      implicit real*8(a-h,o-z)

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     ZK.. - arrays of (1 to M and 1 to N), value for the matrix of
c     coefficient K in nodes
      common/k/ zkxx(200,200), zkxy(200,200),zkyy(200,200)

c     F - array (1 to M-1, and 1 to N-1), values of the right-hand
c     side of the differential equation in cells
      common/f/ f(200,200)


c     =================================================================
c     CC,CE,CNE,CN,CNW,CW,CSW,CS,CSE, PHI - arrays for the coefficients
c                                           and the right-hand side
c     ( 1 to M-1 and 1 to N-1 ) in the finite-difference equations
c
c     C(i,j)*U(i,j)+
c
c     E(i,j)*U(i+1,j) + NE(i,j)*U(i+1,j+1) + N(i,j)*U(i,j+1) +
c
c     NW(i,j)*U(i-1,j+1) + W(i,j)*U(i-1,j) +SW(i,j)*U(i-1,j-1) +
c
c     S(i,j)*U(i,j-1) + SE(i,j)*U(i+1,j-1) = PHI(i,j)
c


c----------------------  STENCIL --------------------------------
c
c
c
c              (i-1,j+2)     (i,j+2)         (i+1,j+2)         (i+2,j+2)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !                 
c             !    NW(i,j)  !   N(i,j)      !   NE(i,j)       !                 
c             !             !               !                 !
c             !             !               !                 ! 
c             !(i-1,j+1)    !(i,j+1)        !(i+1,j+1)        !(i+2,j+1)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !
c             !    W(i,j)   !   C(i,j)      !   E(i,j)        !
c             !             !               !                 !
c             !             !               !                 !
c             !(i-1,j)      !(i,j)          !(i+1,j)          !(i+2,j)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !
c             !    SW(i,j)  !   S(i,j)      !   SE(i,j)       !                 
c             !             !               !                 !
c             !             !               !                 !
c             !(i-1,j-1)    !(i,j-1)        !(i+1,j-1)        !(i+2,j-1)
c             *-------------*---------------*-----------------*
c
c
c
c
c
c----------------------  STENCIL --------------------------------

c
c
      common/coefds/ cc(200,200),ce(200,200),cne(200,200),
     *               cn(200,200),cnw(200,200),cw(200,200),
     *               csw(200,200),cs(200,200),
     *               cse(200,200),phi(200,200)
c=================================================================
c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200,200), VN(200,200)

c     computing the coefficients of the finite-difference scheme
c     and the right-hand side
c     ========================================================

      do  i=1,M-1
      do  j=1,N-1
c
c     COEFFICIENT E(I,J), East, CELL (I+1,J)
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
         cexx = -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(ZKXX(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i+2,j+1)-y(i+1,j))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(-ZKXX(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+1,j+1)-y(i+2,j)))

c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
         cexy = -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(-ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i+2,j+1)-x(i+1,j)) 
     *   -(y(i+1,j+1)-y(i,j))
     *   *(+ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+1,j+1)-x(i+2,j)))
c
c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
         ceyx = -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i+2,j+1)-y(i+1,j))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(-ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+1,j+1)-y(i+2,j)))


c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
         ceyy = -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(-ZKYY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i+2,j+1)-x(i+1,j))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(+ZKYY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+1,j+1)-x(i+2,j)))



      ce(i,j)=0.5*(cexx + cexy + ceyx + ceyy)/VC(i,j)

c
c     COEFFICIENT NE(I,J), North-East, CELL (I+1,J+1)
c
c
      if(i.eq.M-1.and.j.eq.N-1) then
c     There is no cell i+1,j+1 for i=M, j=N
      cne(i,j) = 0.
      goto 4444
      end if
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
      cnexx =
     *    -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(ZKXX(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i+1,j+2)-y(i+2,j+1)))
c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
      cnexy =
     *    -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(-ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i+1,j+2)-x(i+2,j+1)))


c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
      cneyx =
     *    -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i+1,j+2)-y(i+2,j+1)))


c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
      cneyy =
     *    -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(-ZKYY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i+1,j+2)-x(i+2,j+1)))


      cne(i,j)=0.5*(cnexx + cnexy + cneyx + cneyy)/VC(i,j)
4444  continue

c
c     COEFFICIENT N(I,J), North, CELL (I,J+1)
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
         cnxx =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(ZKXX(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i,j+1)-y(i+1,j+2))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(ZKXX(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i,j+2)-y(i+1,j+1)))

c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
         cnxy =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(-ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i,j+1)-x(i+1,j+2))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(-ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i,j+2)-x(i+1,j+1)))
c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
         cnyx =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(ZKXY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (y(i,j+1)-y(i+1,j+2))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i,j+2)-y(i+1,j+1)))
c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
         cnyy =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(-ZKYY(i+1,j+1)/(2.d0*VN(i+1,j+1)))*
     *    (x(i,j+1)-x(i+1,j+2))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(-ZKYY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i,j+2)-x(i+1,j+1)))

      cn(i,j)=0.5*(cnxx + cnxy + cnyx + cnyy)/VC(i,j)


c
c     COEFFICIENT NW(I,J), North-West, CELL (I-1,J+1)
c
c
       if(i.eq.1.and.j.eq.N-1) then
c      There is no cell i-1,j+1 for i=1, j=N
       cnw(i,j) = 0.
       goto 3333
       end if
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
      cnwxx =
     *  -(
     *   - (y(i+1,j+1)-y(i,j))
     *   *(ZKXX(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i-1,j+1)-y(i,j+2)))

c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
      cnwxy =
     *  -(
     *   - (y(i+1,j+1)-y(i,j))
     *   *(-ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i-1,j+1)-x(i,j+2)))


c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
      cnwyx =
     *  -(
     *   + (x(i+1,j+1)-x(i,j))
     *   *(ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i-1,j+1)-y(i,j+2)))


c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
      cnwyy =
     *  -(
     *   + (x(i+1,j+1)-x(i,j))
     *   *(-ZKYY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i-1,j+1)-x(i,j+2)))


      cnw(i,j)=0.5*(cnwxx + cnwxy + cnwyx + cnwyy)/VC(i,j)
3333  continue

c
c     COEFFICIENT W(I,J), West, CELL (I-1,J)
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
         cwxx =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(-ZKXX(i,j)/(2.d0*VN(i,j)))*
     *    (y(i-1,j)-y(i,j+1))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(ZKXX(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i,j)-y(i-1,j+1)))
c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
         cwxy =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(+ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i-1,j)-x(i,j+1))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(-ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i,j)-x(i-1,j+1)))
c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
         cwyx =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(-ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (y(i-1,j)-y(i,j+1))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(ZKXY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (y(i,j)-y(i-1,j+1)))
c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
         cwyy =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(+ZKYY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i-1,j)-x(i,j+1))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(-ZKYY(i,j+1)/(2.d0*VN(i,j+1)))*
     *    (x(i,j)-x(i-1,j+1)))

      cw(i,j)=0.5*(cwxx + cwxy + cwyx + cwyy)/VC(i,j)

c
c     COEFFICIENT SW(I,J), Sowth-West, CELL (I-1,J-1)
c
c
      if(i.eq.1.and.j.eq.1) then 
c     There is no cell (i-1,j-1) for i=1,j=1.
      csw(i,j) = 0.
      goto 1111
      end if
c
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
      cswxx =
     *   -(
     *   +(y(i,j+1)-y(i+1,j))
     *   *(-ZKXX(i,j)/(2.d0*VN(i,j)))*
     *    (y(i,j-1)-y(i-1,j)))

c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
      cswxy =
     *   -(
     *   +(y(i,j+1)-y(i+1,j))
     *   *(+ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i,j-1)-x(i-1,j)))


c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
      cswyx =
     *   -(
     *   -(x(i,j+1)-x(i+1,j))
     *   *(-ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (y(i,j-1)-y(i-1,j)))


c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
      cswyy =
     *   -(
     *   -(x(i,j+1)-x(i+1,j))
     *   *(+ZKYY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i,j-1)-x(i-1,j)))


      csw(i,j)=0.5*(cswxx + cswxy + cswyx + cswyy)/VC(i,j)

1111  continue     
c
c     COEFFICIENT S(I,J), South, CELL (I,J-1)
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
         csxx =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(-ZKXX(i,j)/(2.d0*VN(i,j)))*
     *    (y(i+1,j)-y(i,j-1))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(-ZKXX(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+1,j-1)-y(i,j)))
c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
         csxy =
     *  -(
     *   + (y(i,j+1)-y(i+1,j))
     *   *(+ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i+1,j)-x(i,j-1))
     *   -(y(i+1,j+1)-y(i,j))
     *   *(+ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+1,j-1)-x(i,j)))
c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
         csyx =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(-ZKXY(i,j)/(2.d0*VN(i,j)))*
     *    (y(i+1,j)-y(i,j-1))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(-ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+1,j-1)-y(i,j)))
c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
         csyy =
     *  -(
     *   - (x(i,j+1)-x(i+1,j))
     *   *(+ZKYY(i,j)/(2.d0*VN(i,j)))*
     *    (x(i+1,j)-x(i,j-1))
     *   +(x(i+1,j+1)-x(i,j))
     *   *(+ZKYY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+1,j-1)-x(i,j)))

      cs(i,j)=0.5*(csxx + csxy + csyx + csyy)/VC(i,j)



c
c     COEFFICIENT SE(I,J), South-East, CELL (I+1,J-1)
c
c
       if(i.eq.M-1.and.j.eq.1) then
c      There is no cell i+1,j-1 for i=M,j=1
       cse(i,j) = 0.
       goto 2222
       end if 
c
c    group related to the approximation of term d(zkxx du/dx)/dx
c
      csexx =
     *   -(
     *   -(y(i+1,j+1)-y(i,j))
     *   *(-ZKXX(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+2,j)-y(i+1,j-1)))

c
c    group related to the approximation of term d(zkxy du/dy)/dx
c
      csexy =
     *   -(
     *   -(y(i+1,j+1)-y(i,j))
     *   *(+ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+2,j)-x(i+1,j-1)))


c
c    group related to the approximation of term d(zkxy du/dx)/dy
c
      cseyx =
     *   -(
     *   +(x(i+1,j+1)-x(i,j))
     *   *(-ZKXY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (y(i+2,j)-y(i+1,j-1)))


c
c    group related to the approximation of term d(zkyy du/dy)/dy
c
      cseyy =
     *   -(
     *   +(x(i+1,j+1)-x(i,j))
     *   *(+ZKYY(i+1,j)/(2.d0*VN(i+1,j)))*
     *    (x(i+2,j)-x(i+1,j-1)))


      cse(i,j)=0.5*(csexx + csexy + cseyx + cseyy)/VC(i,j)
2222  continue
c
c      COEFFICIENT IN CELL (I,J) - CC(i,j)


      cc(i,j) = - ( ce(i,j) + cne(i,j) + cn(i,j) + cnw(i,j) +
     *          cw(i,j) + csw(i,j) + cs(i,j) + cse(i,j))


      end do
      end do

      return
      end

c**************************************************************

      subroutine siedel
c
c     SOLUTION for the system of linear algebraic equations
c     by the GAUSS-SIEDEL iteration method
c
      implicit real*8 (a-h,o-z)

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     u - array (0 to M and 0 to N), solution for the finite-difference scheme
c     u at i=0,M and j=0,N - are the given values - Dirichlet boundary
c     conditions
      common/U/ U(0:200,0:200)
c    
c     =================================================================
c     CC,CE,CNE,CN,CNW,CW,CSW,CS,CSE, PHI - arrays for coefficients
c                                           and the right-hand side
c     ( 1 to M-1 and 1 to N-1 ) in the finite-difference equations
c
c     E(i,j)*U(i+1,j) + NE(i,j)*U(i+1,j+1) + N(i,j)*U(i,j+1) +
c
c     NW(i,j)*U(i-1,j+1) + W(i,j)*U(i-1,j) +SW(i,j)*U(i-1,j-1) +
c
c     S(i,j)*U(i,j-1) + SE(i,j)*U(i+1,j-1) = PHI(i,j)
c


c----------------------  STENCIL --------------------------------
c
c
c
c              (i-1,j+2)     (i,j+2)         (i+1,j+2)         (i+2,j+2)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !                 
c             !    NW(i,j)  !   N(i,j)      !   NE(i,j)       !                 
c             !             !               !                 !
c             !             !               !                 ! 
c             !(i-1,j+1)    !(i,j+1)        !(i+1,j+1)        !(i+2,j+1)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !
c             !    W(i,j)   !   C(i,j)      !   E(i,j)        !
c             !             !               !                 !
c             !             !               !                 !
c             !(i-1,j)      !(i,j)          !(i+1,j)          !(i+2,j)
c             *-------------*---------------*-----------------*
c             !             !               !                 !
c             !             !               !                 !
c             !             !               !                 !
c             !    SW(i,j)  !   S(i,j)      !   SE(i,j)       !                 
c             !             !               !                 !
c             !             !               !                 !
c             !(i-1,j-1)    !(i,j-1)        !(i+1,j-1)        !(i+2,j-1)
c             *-------------*---------------*-----------------*
c
c
c
c
c
c----------------------  STENCIL --------------------------------

c
c

c
c
      common/coefds/ cc(200,200),ce(200,200),cne(200,200),
     *               cn(200,200),cnw(200,200),cw(200,200),
     *               csw(200,200),cs(200,200),
     *               cse(200,200),phi(200,200)
c=================================================================

c     psie,psin,psiw and psis 1-dimensional arrays
c     (1 to M-1,  and 1 to N-1, correspondingly to part of the boundary),
c     -  values of the function of the right-hand side in the Dirichlet 
c     boundary conditions.
      common/psi/ psie(200),psin(200),psiw(200),psis(200)

c=================================================================

c     eps - constant in the condition of finishing the iteration in 
c     the Gauss-Siedel iteration procedure.  The termination condition is
c
c             (S+1)     (S)
c     max( abs( U(i,j) - U(i,j))) < eps
c     i,j
c 
c     where S is the iteration number.
c

      common/eps/ eps
c=====================================================================
c     SET an INITIAL value for the iteration solution
c     North and South Boundary
      do  i=1,M-1
      U(i,N)=psin(i)
      U(i,0)=psis(i)
      end do

c     East and West Boundary
      do  j=1,N-1
      U(M,j)=psie(j)
      U(0,j)=psiw(j)
      end do

c     Beginning of the iteration process
c
5      continue

       EPSMAX= 0.D0
c
       do  i=1, M-1
       do  j=1, N-1
       USP1=( phi(i,j)-
     * (CE(i,j)*U(i+1,j) + CNE(i,j)*U(i+1,j+1) + CN(i,j)*U(i,j+1) +
     * CNW(i,j)*U(i-1,j+1) + CW(i,j)*U(i-1,j) +CSW(i,j)*U(i-1,j-1) +
     * CS(i,j)*U(i,j-1) + CSE(i,j)*U(i+1,j-1)))/CC(i,j)
       EPSMAX=amax1(EPSMAX,dabs(USP1 - U(i,j)))
       U(i,j)=USP1
       end do
       end do
c
c     checking the termination condition
       if(EPSMAX.GT.EPS) goto 5

       return
       end
c*******************************************************************
      
       subroutine mkgrid
c      This is the subroutine for generation of a smooth grid in 2D
c      as shown in the book.
c
c      Left boundary x=0,
c      Right boundary x=1,
c      Bottom boundary y=0,
c      Upper boundary y(x)=1.+0.5*sin(2.*pi*x)
c           _
c          / \
c         /   \
c        /     \
c       /       \
c      /         \
c      I          \                   /I
c      I           \                 / I
c      I            \               /  I
c      I             \             /   I
c      I              \           /    I
c      I               \         /     I
c      I                \       /      I
c      I                 \     /       I
c      I                  \   /        I
c      I                   \_/         I
c      I                               I
c      I_______________________________I
c
c
c      
c 

      implicit real*8 (a-h,o-z)

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

      M=17

      N=17

      pi=acos(-1.)

      hx=1./(M-1.)


      do i=1, M
      do j=1, N
      xi=(i-1)*hx
      et=(j-1.)/(M-1.)  
      hyj=(1.+0.5*sin(2.*pi*xi))/(M-1.)
      x(i,j)=xi
      y(i,j)=(j-1.)*hyj 
      end do
      end do

c     Fictitious values of X and Y

      do i=1,M
      x(i,N+1) = x(i,N)
      y(i,N+1) = y(i,N)
      x(i,0) = x(i,1)
      y(i,0) = y(i,1)
      end do

      do j=1,N
      x(M+1,j) = x(M,j)
      y(M+1,j) = y(M,j)
      x(0,j) = x(1,j)
      y(0,j) = y(1,j)
      end do

      return
      end
c**********************************************************************

      subroutine exact
c     Compute the exact solution for the test problem
c     (Dirichlet boundary problem for the Poisson equation)
c     in nodes
c     Results in array
c     uex - 1 to M and 1 to N

      implicit real*8 (a-h,o-z)
 
c     t - current time
      common/t/ t,dt
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     UEX - array 0 to M and 0 to N; Exact solution
      common/uex/ uex(0:200,0:200)

      pi=acos(-1.)

      do i=1,M-1
      do j=1,N-1
      xcij=0.25*(x(i,j)+x(i+1,j)+x(i+1,j+1)+x(i,j+1))
      ycij=0.25*(y(i,j)+y(i+1,j)+y(i+1,j+1)+y(i,j+1))
 
      uex(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)

      end do
      end do

      i=0
      do j=1,N-1
      xcij=0.5*(x(i,j)+x(i,j+1))
      ycij=0.5*(y(i,j)+y(i,j+1))
      uex(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      i=M
      do j=1,N-1
      xcij=0.5*(x(i,j)+x(i,j+1))
      ycij=0.5*(y(i,j)+y(i,j+1))
      uex(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      j=0
      do i=1,M-1
      xcij=0.5*(x(i,j)+x(i+1,j))
      ycij=0.5*(y(i,j)+y(i+1,j))
      uex(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      j=N
      do i=1,M-1
      xcij=0.5*(x(i,j)+x(i+1,j))
      ycij=0.5*(y(i,j)+y(i+1,j))
      uex(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      return
      end
c*****************************************************************

      subroutine output

c     Output subroutine 
c     prints the approximate solution,
c     exact solution and errors for
c     the test problem

      implicit real*8 (a-h,o-z)
 
c     t - current time
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     UEX - array 0 to M and 0 to N; Exact solution
      common/uex/ uex(0:200,0:200)

c     u - array (0 to M and 0 to N), solution for the finite-difference scheme
c     u at i=0,M and j=0,N - are the given values - Dirichlet boundary
c     conditions
      common/U/ U(0:200,0:200)

c     ERR - array 0 to M and 0 to N; Error between the exact and
c           approximate solutions in nodes
      common/err/ err(0:200,0:200)

c     Max-norm and L2-norm of the error between the exact and approximate
c     solutions in nodes
      common/errors/supn,zL2N

c     Open the file h2ddcn.output
c     for test output of program h2ddcn
      open(unit=1,file='h2ddcn.output')
      
      write(1,33) 
33    format(15X,'TEST OUTPUT FOR PROGRAM h2ddcn')
      write(1,*)
      if(sigma.eq.0) then
      write(1,*)'               Explicit Scheme'
      else
      write(1,*)'               Implicit Scheme',' Sigma= ', sigma
      end if
      write(1,*)
 
 
      write(1,8) t,dt, M,N
      write(1,*)
 
 
      write(1,88) M,N, supn, zL2N
8     format(2X,'Time= ',E12.5,2X,'dt = ',
     * E12.5,2X, ' Number of nodes M,N = ',I5,2X,I5)
88    format(2X,'M=',I5,2X,'N=',I5,2X,
     *       'MAX-NORM = ',E12.5,2X,'  L2-NORM = ',E12.5)

      
       write(1,*)

c     Print of the approximate and exact solutions of the test problem
c     Print for I=I1,I2,I3 and for all J=0,...,N

      I1=2
      I2=9
      I3=16
      write(1,9) I1
9     format(10X,'Approximate and Exact Solutions, and Error for',
     * ' I=',I5)
       write(1,*)
      
      write(1,10)
10    format(5X,'j',7X,' X',12X,' Y',10X,'Appr.Sol.',
     * 4X,'Exact Sol.',5X,'Error')
       i=I1
      do j=0,N 
      write(1,11) j,x(i,j),y(i,j),u(i,j),uex(i,j),err(i,j)
      end do
       write(1,*)
       write(1,*)
11    format(1X, i5,5(2X,E12.5))

      write(1,9) I2
       write(1,*)
       i=I2
      do j=0,N 
      write(1,11) j,x(i,j),y(i,j),u(i,j),uex(i,j),err(i,j)
      end do
       write(1,*)
       write(1,*)

      write(1,9) I3
       write(1,*)
       i=I3
      do j=0,N 
      write(1,11) j,x(i,j),y(i,j),u(i,j),uex(i,j),err(i,j)
      end do
       write(1,*)
       write(1,*)

      return
      end
c*********************************************************************

      subroutine error
c     compute the Max and L2 norm of
c     error for the test problem
      implicit real*8 (a-h,o-z)

c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200,200), VN(200,200)
 
c     Comments - this volume is needed in this subroutine
c     only for computing the L2 norm of error
 
c     t - current time
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     UEX - array 0 to M and 0 to N; Exact solution
      common/uex/ uex(0:200,0:200)

c     u - array (0 to M and 0 to N), solution for the finite-difference scheme
c     u at i=0,M and j=0,N - are the given values - Dirichlet boundary
c     conditions
      common/U/ U(0:200,0:200)

c     ERR - array 0 to M and 0 to N; Error between the exact and
c           approximate solutions in nodes
      common/err/ err(0:200,0:200)

c     Max-norm and L2-norm of the error between the exact and approximate
c     solutions in nodes
      common/errors/supn,zL2N

 
c     Comments - this volume is needed in this subroutine
c     only for computing the L2 norm of error

c     Computing the exact solution
      call exact
c     
c     Error
c     -----

      do i=1,M-1
      do j=1,N-1
      err(i,j)=uex(i,j)-u(i,j)
      end do
      end do

c     Max-Norm
c     --------
      supn=0.d0
      do i=1,M-1
      do j=1,N-1
      supn=amax1(supn,abs(err(i,j)))
      end do
      end do
      
c     L  - norm
c      2
c     ----------

      zL2N=0.0d0
      do i=1,M-1
      do j=1,N-1
c     For the Dirichlet boundary problem, error on the boundary is equal
c     to zero and therefore, we do not compute terms in the inner
c     product that are related to boundary i=0;M j=0;N.
      zL2N = zL2N + err(i,j)**2*VC(i,j)
      end do
      end do
      zL2N=dsqrt(zL2N)
      
      return
      end
c**********************************************************************

      subroutine bcon
c     This subroutine computes the values of the right-
c     hand side of the Dirichlet boundary conditions
c     in a finite-difference scheme.
c     fpsin(x,y),fpsis(x,y),fpsie(x,y),fpsiw(x,y)
c     functions which compute the right-hand side
c     in the Dirichlet boundary conditions on
c     related parts of the boundary  
     
      implicit real*8 (a-h,o-z)
c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)
   
c     psie,psin,psiw and psis 1-dimensional arrays
c     (1 to M,  and 1 to N, correspondingly to part of the boundary),
c     -  values of the function of the right-hand side in the Dirichlet 
c     boundary conditions.
      common/psi/ psie(200),psin(200),psiw(200),psis(200)

c     Set the Dirichlet boundary conditions
c     --------------------------------------

c     North and South Boundary
      do  i=1,M-1
      xciN=0.5*(x(i,N)+x(i+1,N))
      yciN=0.5*(y(i,N)+y(i+1,N))
      psin(i) = fpsin(xciN,yciN)

      xci1=0.5*(x(i,1)+x(i+1,1))
      yci1=0.5*(y(i,1)+y(i+1,1))
      psis(i) = fpsis(xci1,yci1)
      end do

c     East and West Boundary
      do  j=1,N-1
      xcMj=0.5*(x(M,j)+x(M,j+1))
      ycMj=0.5*(y(M,j)+y(M,j+1))
      psie(j) = fpsie(xcMj,ycMj)

      xc1j=0.5*(x(1,j)+x(1,j+1))
      yc1j=0.5*(y(1,j)+y(1,j+1))
      psiw(j) = fpsiw(xc1j,yc1j)
      end do

      return
      end
c***********************************************************************

      real function fpsin*8(x,y)
c     function fpsin - returned values for the right-hand
c     side of the Dirichlet boundary conditions of
c     a differential problem
c     North - Top - boundary -> j=N
      implicit real*8 (a-h,o-z)
 
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
 
 
      tt=t

      pi=acos(-1.)
      fpsin=exp(-2.*pi**2*tt)*sin(pi*x)*sin(pi*y)
      return
      end
c***********************************************************************

      real function fpsis*8(x,y)
c     function fpsin - returned values for the right-hand
c     side of the Dirichlet boundary conditions of
c     a differential problem
c     South - Bottom - boundary -> j=1
      implicit real*8 (a-h,o-z)
 
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
 
 
      tt=t

      pi=acos(-1.)
      fpsis=exp(-2.*pi**2*tt)*sin(pi*x)*sin(pi*y)
      return
      end
c***********************************************************************

      real function fpsie*8(x,y)
c     function fpsin - returned values for the right-hand
c     side of the Dirichlet boundary conditions of
c     a differential problem
c     East - Right - boundary -> i=M
      implicIt real*8 (a-h,o-z)
 
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
 
 
      tt=t

      pi=acos(-1.)
      fpsie=exp(-2.*pi**2*tt)*sin(pi*x)*sin(pi*y)
      return
      end
c***********************************************************************

      real function fpsiw*8(x,y)
c     function fpsin - returned values for the right-hand
c     side of the Dirichlet boundary conditions of
c     a differential problem
c     West - left - boundary -> i=1
      implicit real*8 (a-h,o-z)
 
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
 
 
      tt=t

      pi=acos(-1.)
      fpsiw=exp(-2.*pi**2*tt)*sin(pi*x)*sin(pi*y)
      return
      end
c***********************************************************************

      subroutine coeff
c     Computation of coefficients K.. of the finite-difference
c     scheme in cells
      implicit real*8 (a-h,o-z)
c     M,N - number of nodes in the grid 
      common/MN/ M,N
 
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma
 

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     ZK.. - arrays of (1 to M and 1 to N), value of the matrix of
c     coefficient K in nodes
      common/k/ zkxx(200,200), zkxy(200,200),zkyy(200,200)

      tt=t-0.5*dt

      do i=1,M
      do j=1,N


c     computation of coefficients

      zkxx(i,j)=fkxx(x(i,j),y(i,j),tt)                                     
      zkxy(i,j)=fkxy(x(i,j),y(i,j),tt)
      zkyy(i,j)=fkyy(x(i,j),y(i,j),tt)
      
      end do
      end do

      return
      end
c*********************************************************************

      real function fkxx*8 (x,y,tt)
c     function fkxx(x,y) computes the coefficients
c     KXX of the original differential equation in point (x,y)
      implicit real*8 (a-h,o-z)
      pi=acos(-1.)
      d1 = 1.0d0 + 2.0d0*x**2 + y**2
      d2 = 1.0d0 + x**2 + 2.0d0*y**2
      t = 5.d0*pi/12.d0
      ct=cos(t)
      st=sin(t)
      fkxx=d1*ct*ct + d2*st*st
      return
      end
c*********************************************************************

      real function fkxy*8 (x,y,tt)
c     function fkxy(x,y) computes the coefficients
c     KXX of the original differential equation in point (x,y)
      implicit real*8 (a-h,o-z)
      pi=acos(-1.)
      d1 = 1.0d0 + 2.0d0*x**2 + y**2
      d2 = 1.0d0 + x**2 + 2.0d0*y**2
      t = 5.d0*pi/12.d0
      ct=cos(t)
      st=sin(t)
      fkxy=(d1-d2)*ct*st
      return
      end
c*********************************************************************

      real function fkyy*8 (x,y,tt)
c     function fkxy(x,y) computes the coefficients
c     KXX of the original differential equation in point (x,y)
      implicit real*8 (a-h,o-z)
      pi=acos(-1.)
      d1 = 1.0d0 + 2.0d0*x**2 + y**2
      d2 = 1.0d0 + x**2 + 2.0d0*y**2
      t = 5.d0*pi/12.d0
      ct=cos(t)
      st=sin(t)
      fkyy=d1*st*st + d2*ct*ct
      return
      end
c*********************************************************************

      subroutine right
c     Computation of the right-hand side of the finite-difference 
c     equations in nodes
      implicit real*8 (a-h,o-z)

c     t - current time
      common/t/ t,dt
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma


c     M,N - number of nodes in the grid 
      common/MN/ M,N

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)

c     F - array (1 to M-1, and 1 to N-1), values of the right-hand
c     side of the differential equation in cells
      common/f/ f(200,200)

c     comuting the right-hand side of the finite-difference equations
c     ff(x,y) - function which computes the value of the right-hand side
c     of the original differential problem

      tt=t-0.5*dt


      do i=1,M-1
      do j=1,N-1
      xcij=0.25*(x(i,j)+x(i+1,j)+x(i+1,j+1)+x(i,j+1))
      ycij=0.25*(y(i,j)+y(i+1,j)+y(i+1,j+1)+y(i,j+1))
      f(i,j)=ff(xcij,ycij,tt)
      end do
      end do

      return
      end
c**********************************************************************

      real function ff*8(x,y,tt)
c     function ff(x,y,tt) - computes the value of the right-hand side
c     of the original differential equation in point (x,y)
c     at moment tt

      implicit real*8 (a-h,o-z)
      pi=acos(-1.)
     
            d1  = 1.0d0 + 2.0d0*x**2 + y**2
            d2  = 1.0d0 + x**2 + 2.0d0*y**2
            d1x = 4.0d0*x
            d1y = 2.0d0*y
            d2x = 2.0d0*x
            d2y = 4.0d0*y
            t  = 5.d0*pi/12.d0
            ct = dcos(t)
            st = dsin(t)
            zkxx = d1*ct*ct + d2*st*st
            zkxy = (d1-d2)*ct*st
            zkyy = d1*st*st + d2*ct*ct
            zkxx_x     = d1x*ct*ct + d2x*st*st
            zkxy_x    = ( d1x - d2x )*ct*st
            zkxy_y    = ( d1y - d2y )*ct*st
            zkyy_y     = d1y*st*st + d2y*ct*ct

c
            sx  = sin(pi*x)
            cx  = cos(pi*x)
            sy  = sin(pi*y)
            cy  = cos(pi*y)
            fx       =     pi*cx*sy
            fy       =     pi*sx*cy
            fxx      = -pi*pi*sx*sy
            fxy      =  pi*pi*cx*cy
            fyy      = -pi*pi*sx*sy
c      
            ff = -
     *  exp(-2.*pi**2*tt)
c    *  exp(-pi**2*tt)
     *   *((zkxx*fxx + 2*zkxy*fxy + zkyy*fyy
     *  + zkxx_x*fx + zkxy_y*fx + zkxy_x*fy + zkyy_y*fy)
     *  +2.*pi**2*sx*sy)
c    *  +pi**2*sx)
 
      return
      end
c************************************************************************
      subroutine geometry
      implicit real*8 (a-h,o-z)
 
c     M,N - number of nodes in the grid
      common/MN/ M,N
 
c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)
 
c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200,200), VN(200,200)
 
c     Computing VC and VN
 
      do  i=1,M-1
      do  j=1,N-1
      VC(i,j) =0.5*((x(i,j)-x(i+1,j+1))*(y(i+1,j)-y(i,j+1)) -
     *              (x(i+1,j)-x(i,j+1))*(y(i,j)-y(i+1,j+1)))
 
      end do
      end do
 
 
      do  i=1,M
      do  j=1,N
c     Two different choices for VN(i,j)
c     First - average of volumes of the adjacent cells
c     VN(i,j)=0.25*(VC(i,j) + VC(i-1,j) + VC(i-1,j-1) + VC(i,j-1))
c     Second "DIAMOND" volume
      VN(i,j)=0.25*((x(i-1,j)-x(i+1,j))*(y(i,j-1)-y(i,j+1)) -
     *              (x(i,j-1)-x(i,j+1))*(y(i-1,j)-y(i+1,j)))
 
      end do
      end do
 
      return
      end
c**************************************************************************
      subroutine initial
 
      implicit real*8 (a-h,o-z)
 
c     M,N - number of nodes in the grid
      common/MN/ M,N
 
c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)
 
c     u - array (0 to M and 0 to N), solution for the finite-difference scheme
c     u at i=0,M and j=0,N - are the given values - Dirichlet boundary
c     conditions
      common/U/ U(0:200,0:200)
 
      pi=acos(-1.)
 
c     Set initial conditions at t=0.
 
      do i=1,M-1
      do j=1,N-1
      xcij=0.25*(x(i,j)+x(i+1,j)+x(i+1,j+1)+x(i,j+1))
      ycij=0.25*(y(i,j)+y(i+1,j)+y(i+1,j+1)+y(i,j+1))
 
      u(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)

      end do
      end do

      i=0
      do j=1,N-1
      xcij=0.5*(x(i,j)+x(i,j+1))
      ycij=0.5*(y(i,j)+y(i,j+1))
      u(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      i=M
      do j=1,N-1
      xcij=0.5*(x(i,j)+x(i,j+1))
      ycij=0.5*(y(i,j)+y(i,j+1))
      u(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      j=0
      do i=1,M-1
      xcij=0.5*(x(i,j)+x(i+1,j))
      ycij=0.5*(y(i,j)+y(i+1,j))
      u(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do

      j=N
      do i=1,M-1
      xcij=0.5*(x(i,j)+x(i+1,j))
      ycij=0.5*(y(i,j)+y(i+1,j))
      u(i,j)=
     *   exp(-2.*pi**2*t)*sin(pi*xcij)*sin(pi*ycij)
      end do
 
      return
      end  
c**********************************************************
      subroutine hcoeff
c      This is subroutine hcoeff for computation of the 
c      coefficient and the right-hand side
c      of the finite-difference scheme for the implicit FDS, 
c      and approximation of the Robin boundary conditions.
c=================================================================
c     Discretization - SCALAR functions in NODES (HN),
c                    - VECTOR functions in CELLS (HC)
c=================================================================
      implicit real*8(a-h,o-z)
 
c     t - current time
      common/t/ t,dt
 
c     M,N - number of nodes in the grid
      common/MN/ M,N
 
c     sigma - parameter in the finite-difference scheme
      common/expimp/ sigma

c     X,Y - arrays (0 to M+1 and 0 to N+1), coordinates of grid nodes
c     X and Y for i=0,M+1; j=0,N+1 are fictitious, and are used
c     for making the computational formulas and program code shorter.
      common/xy/ x(0:201,0:201),y(0:201,0:201)
 
c     u - array (0 to M and 0 to N), solution for the finite-difference scheme
c     u at i=0,M and j=0,N - are the given values - Dirichlet boundary
c     conditions
      common/U/ U(0:200,0:200)
 
      common/coefds/ cc(200,200),ce(200,200),cne(200,200),
     *               cn(200,200),cnw(200,200),cw(200,200),
     *               csw(200,200),cs(200,200),
     *               cse(200,200),phi(200,200)

c     F - array (1 to M-1, and 1 to N-1), values for the right-hand
c     side of the differential equation in cells
      common/f/ f(200,200)
 
      do  i=1,M-1
      do  j=1,N-1
 
      phi(i,j)=phi(i,j)
     *-(1.-sigma)*(cc(i,j)*U(i,j)+
     *ce(i,j)*U(i+1,j) + cne(i,j)*U(i+1,j+1) + cn(i,j)*U(i,j+1) +
     *cnw(i,j)*U(i-1,j+1)+cw(i,j)*U(i-1,j)+csw(i,j)*U(i-1,j-1) +
     *cs(i,j)*U(i,j-1) +cse(i,j)*U(i+1,j-1))
 
       cc(i,j) =
     * sigma*cc(i,j)
       ce(i,j) =
     * sigma*ce(i,j)
       cne(i,j) =
     * sigma*cne(i,j)
       cn(i,j) =
     * sigma*cn(i,j)
       cnw(i,j) =
     * sigma*cnw(i,j)
       cw(i,j) =
     * sigma*cw(i,j)
       csw(i,j) =
     * sigma*csw(i,j)
       cs(i,j)  =
     * sigma*cs(i,j)
       cse(i,j) =
     * sigma*cse(i,j)
 
       end do
       end do
 
 
      do  i=1,M-1
      do  j=1,N-1
 
      phi(i,j)=phi(i,j)+U(i,j)/dt
      cc(i,j)=cc(i,j) + 1./dt
 
       end do
       end do
 
         return
         end
c*****************************************************************
      subroutine sphi
      implicit real*8(a-h,o-z)
 
c     M,N - number of nodes in the grid
      common/MN/ M,N
 
c     F - array (1 to M-1, and 1 to N-1), values for the right-hand
c     side of the differential equation in cells
      common/f/ f(200,200)
 
c
      common/coefds/
     *      cc(200,200),ce(200,200),
     *      cne(200,200),cn(200,200),
     *      cnw(200,200),cw(200,200),
     *      csw(200,200),cs(200,200),
     *      cse(200,200),phi(200,200)
c=================================================================

      do i=1,M-1
      do j=1,N-1
      phi(i,j)=f(i,j)
      end do
      end do
 
      return
      end
 

c************************************************************************
      subroutine step
c     Explicit finite-difference scheme
c     for the Heat equation with Dirichlet boundary conditions
c
c      du/dt = [d(KXX du/dx)/dx+d(KXY du/dy)/dx
c              +d(KXY du/dx)/dy+d(KYY du/dy)/dy+f
c
c             with boundary conditions
c             u(x,y,t)=psi(x,y,t) for (x,y) on the boundary,
c             and intial  conditions
c             u(x,y,0)=u_0(x,y) at t=0.
c      for a 2-D case.
c=================================================================
c     Discretization - SCALAR functions in CELLS (HC),
c                    - VECTOR functions in NODES (HN)
c=================================================================
      implicit real*8(a-h,o-z)
c     t - current time
      common/t/ t,dt
 
c     M,N - number of nodes in the grid
      common/MN/ M,N
 
 
c     u - array (0 to M and 0 to N), solution on the previous time level
c     unp1 - array (0 to M and 0 to N), solution on the new time level
      common/U/ U(0:200,0:200)
      dimension unp1(0:200,0:200)
 
c     X,Y - arrays (1 to M and 1 to N), coordinates of grid nodes
      common/xy/ x(0:201,0:201),y(0:201,0:201)
 
c     ZK.. - arrays of (1 to M and 1 to N), value of the matrix of
c     coefficient K in nodes
      common/k/ zkxx(200,200), zkxy(200,200),zkyy(200,200)
 
c     F - array (1 to M-1, and 1 to N-1), values of the right-hand
c     side of the differential equation in cells
      common/f/ f(200,200)
 
c     psie,psin,psiw and psis 1-dimensional arrays
c     (1 to M-1,  and 1 to N-1, correspondingly to part of the boundary),
c     -  values of the function of the right-hand side in the Dirichlet 
c     boundary conditions.
      common/psi/ psie(200),psin(200),psiw(200),psis(200)
 
c     sigma - parameter in the finite-difference scheme
 
      common/expimp/ sigma
 
 
 
c     =================================================================
c     CC,CE,CNE,CN,CNW,CW,CSW,CS,CSE, PHI - arrays for the coefficients
c                                           and the right-hand side
c     ( 2 to M-1 and 2 to N-1 ) in the finite-difference equations
c
c     C(i,j)*U(i,j)+
c
c     E(i,j)*U(i+1,j) + NE(i,j)*U(i+1,j+1) + N(i,j)*U(i,j+1) +
c
c     NW(i,j)*U(i-1,j+1) + W(i,j)*U(i-1,j) +SW(i,j)*U(i-1,j-1) +
c
c     S(i,j)*U(i,j-1) + SE(i,j)*U(i+1,j-1) = PHI(i,j)
c
 
 
c----------------------  STENCIL --------------------------------
c
c
c
 
c
c              NW(i,j)       N(i,j)          NE(i,j)
c             *-------------*---------------*
c             !             !               !
c             !             !               !
 
 
c             !             !               !
c             !             !               !
c             !             !               !
c             !W(i,j)       !C(i,j)         !E(i,j)
c             *-------------*---------------*
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !             !               !
c             !SW(i,j)      !S(i,j)         !SE(i,j)
 
c             *-------------*---------------*
c
c
c
c
c
c----------------------  STENCIL --------------------------------
 
 
c
      common/coefds/ cc(200,200),ce(200,200),cne(200,200),
     *               cn(200,200),cnw(200,200),cw(200,200),
     *               csw(200,200),cs(200,200),
     *               cse(200,200),phi(200,200)
c=================================================================
 
c     VC, VN - arrays for the volume of cells and nodes
      common/volumes/ VC(200,200), VN(200,200)
 
c     computing the coefficients of the finite-difference scheme
c     and the right-hand side
c     ========================================================
 
 
c     North and South Boundary
      do  i=1,M-1
      unp1(i,N)=psin(i)
      unp1(i,0)=psis(i)
      end do
 
c     East and West Boundary
      do  j=1,N-1
      unp1(M,j)=psie(j)
      unp1(0,j)=psiw(j)
      end do
 
 
      do  i=1,M-1
      do  j=1,N-1
 
      unp1(i,j)= U(i,j) +
     * dt*(phi(i,j)
     *-(cc(i,j)*U(i,j)+
     *ce(i,j)*U(i+1,j) + cne(i,j)*U(i+1,j+1) + cn(i,j)*U(i,j+1) +
     *cnw(i,j)*U(i-1,j+1)+cw(i,j)*U(i-1,j)+csw(i,j)*U(i-1,j-1) +
     *cs(i,j)*U(i,j-1) +cse(i,j)*U(i+1,j-1)))
 
 
 
 
       end do
       end do
 
c      Overdetermination of U
c      because we use only two arrays.
c      If you need to use both arrays, do 
c      it before this overdetermination
c      and put this loop in another subroutine
 
 
      do  i=0,M
      do  j=0,N
 
      U(i,j)=unp1(i,j)
 
      end do
      end do
 
      return
      end
 
c *****************************************************************



